part of pdf;

/// Provides methods and properties to create PDF pages and its elements, [PdfPage].
class PdfPage implements _IPdfWrapper {
  //Constructor
  /// Initializes a new instance of the [PdfPage] class.
  PdfPage() {
    _initialize();
  }

  //Fields
  bool _isNewPage;
  PdfSection _section;
  PdfAnnotationCollection _annotations;
  _PdfDictionary _dictionary;
  _PdfResources _resources;
  PdfPageLayerCollection _layers;
  int _defaultLayerIndex = -1;

  /// Raises before the page saves.
  Function _beginSave;

  //Properties
  /// Gets size of the PDF page- Read only
  Size get size => _section.pageSettings.size;
  Offset get _origin => _section.pageSettings._origin.offset;
  PdfDocument get _document {
    if (_section != null && _section._parent != null) {
      return _section._parent._document;
    } else {
      return null;
    }
  }

  /// Gets a collection of the annotations of the page- Read only.
  PdfAnnotationCollection get annotations {
    if (_annotations == null) {
      _annotations = PdfAnnotationCollection(this);
      if (!_dictionary.containsKey(_DictionaryProperties.annots)) {
        _dictionary[_DictionaryProperties.annots] =
            _annotations._internalAnnotations;
      }
      _annotations._internalAnnotations =
          _dictionary[_DictionaryProperties.annots] as _PdfArray;
    }
    return _annotations;
  }

  /// Gets the graphics of the [DefaultLayer].
  PdfGraphics get graphics => defaultLayer.graphics;

  /// Gets the collection of the page's layers (Read only).
  PdfPageLayerCollection get layers {
    _layers ??= PdfPageLayerCollection(this);
    return _layers;
  }

  /// Gets array of page's content.
  _PdfArray get _contents {
    final _IPdfPrimitive object = _dictionary[_DictionaryProperties.contents];
    _PdfArray elements = object as _PdfArray;
    if (object is _PdfReferenceHolder) {
      final _PdfReferenceHolder holder = object;
      elements = holder.object as _PdfArray;
      if (elements == null) {
        final _PdfStream contentStream = holder.object as _PdfStream;
        if (contentStream != null) {
          elements = _PdfArray();
          elements._add(_PdfReferenceHolder(contentStream));
          _dictionary[_DictionaryProperties.contents] = elements;
        }
      }
    }
    if (elements == null) {
      elements = _PdfArray();
      _dictionary[_DictionaryProperties.contents] = elements;
    }
    return elements;
  }

  /// Gets the default layer of the page (Read only).
  PdfPageLayer get defaultLayer => layers[defaultLayerIndex];

  /// Gets or sets index of the default layer (Read only).
  int get defaultLayerIndex {
    if (layers.count == 0 || _defaultLayerIndex == -1) {
      final PdfPageLayer layer = layers.add();
      _defaultLayerIndex = layers.indexOf(layer);
    }
    return _defaultLayerIndex;
  }

  PdfPageOrientation get _orientation => _obtainOrientation();

  //Public methods
  Size getClientSize() {
    return _section._getActualBounds(this, true).size.size;
  }

  //Implementation

  void _assignSection(PdfSection section) {
    ArgumentError.checkNotNull(section, 'section');
    _section = section;
    _dictionary[_DictionaryProperties.parent] = _PdfReferenceHolder(section);
  }

  void _initialize() {
    _dictionary = _PdfDictionary();
    _dictionary[_DictionaryProperties.type] = _PdfName('Page');
    _dictionary._beginSave = _pageBeginSave;
  }

  void _drawPageTemplates(PdfDocument document) {
    // Draw Background templates.
    final bool hasBackTemplates =
        _section._containsTemplates(document, this, false);
    if (hasBackTemplates) {
      final PdfPageLayer backLayer =
          PdfPageLayer._fromClipPageTemplate(this, false);
      final PdfPageLayerCollection _layer = PdfPageLayerCollection(this);
      _layers = _layer;
      _layers.addLayer(backLayer);
      _section._drawTemplates(this, backLayer, document, false);
    }

    // Draw Foreground templates.
    final bool hasFrontTemplates =
        _section._containsTemplates(document, this, true);

    if (hasFrontTemplates) {
      final PdfPageLayer frontLayer =
          PdfPageLayer._fromClipPageTemplate(this, false);
      final PdfPageLayerCollection _layer = PdfPageLayerCollection(this);
      _layers = _layer;
      _layers.addLayer(frontLayer);
      _section._drawTemplates(this, frontLayer, document, true);
    }
  }

  PdfPageOrientation _obtainOrientation() {
    return (size.width > size.height)
        ? PdfPageOrientation.landscape
        : PdfPageOrientation.portrait;
  }

  _PdfResources _getResources() {
    if (_resources == null) {
      _resources = _PdfResources();
      _dictionary[_DictionaryProperties.resources] = _resources;
    }
    return _resources;
  }

  void _setResources(_PdfResources resources) {
    _resources = resources;
    _dictionary[_DictionaryProperties.resources] = _resources;
  }

  void _pageBeginSave(Object sender, _SavePdfPrimitiveArgs args) {
    final PdfDocument doc = args.writer._document;
    if (doc != null && _document != null) {
      _drawPageTemplates(doc);
    }
    if (_beginSave != null) {
      _beginSave();
    }
  }

  //_IPdfWrapper elements
  @override
  _IPdfPrimitive get _element => _dictionary;
  @override
  //ignore: unused_element
  set _element(_IPdfPrimitive value) {
    _dictionary = value;
  }
}
