part of pdf;

/// This class plays two roles: it's a base class for all bookmarks
/// and it's a root of a bookmarks tree.
class PdfBookmarkBase implements _IPdfWrapper {
  /// Initializes a new instance of the [PdfBookmarkBase] class.
  PdfBookmarkBase._internal() : super();

  /// Collection of the descend outlines.
  final List<PdfBookmarkBase> _list = <PdfBookmarkBase>[];

  /// Internal variable to store loaded bookmark.
  List<PdfBookmark> _bookmark;

  /// Internal variable to store dictinary.
  final _PdfDictionary _dictionary = _PdfDictionary();

  /// Whether the bookmark tree is expanded or not
  bool _expanded = false;

  /// Gets number of the elements in the collection. Read-Only.
  int get count => _list.length;

  /// Gets the bookmark at specified index.
  PdfBookmark operator [](int index) {
    return _list[index] as PdfBookmark;
  }

  /// Gets whether to expand the node or not
  bool get _isExpanded {
    if (_dictionary.containsKey('Count')) {
      return (_dictionary[_DictionaryProperties.count] as _PdfNumber).value < 0
          ? false
          : true;
    } else {
      return _expanded;
    }
  }

  /// Sets whether to expand the node or not
  set _isExpanded(bool value) {
    _expanded = value;
    if (count > 0) {
      final int newCount = _expanded ? _list.length : -_list.length;
      _dictionary[_DictionaryProperties.count] = _PdfNumber(newCount);
    }
  }

  /// Adds the bookmark from the document.
  PdfBookmark add(String title) {
    ArgumentError.checkNotNull(title, 'title');
    final PdfBookmark previous = (count < 1) ? null : this[count - 1];
    final PdfBookmark outline =
        PdfBookmark._internal(title, this, previous, null);
    if (previous != null) {
      previous._next = outline;
    }
    _list.add(outline);
    _updateFields();
    return outline;
  }

  /// Determines whether the specified outline presents in the collection.
  bool contains(PdfBookmark outline) {
    return _list.contains(outline);
  }

  /// Remove the specified bookmark from the document.
  void remove(String title) {
    ArgumentError.checkNotNull(title, 'title');
    int index = -1;
    if (_bookmark == null || _bookmark.length < _list.length) {
      _bookmark = <PdfBookmark>[];
      for (int n = 0; n < _list.length; n++) {
        ArgumentError.checkNotNull(_list[n] as PdfBookmark, 'bookmark');
        _bookmark.add(_list[n] as PdfBookmark);
      }
    }
    for (int c = 0; c < _bookmark.length; c++) {
      if (_bookmark[c] is PdfBookmark) {
        final PdfBookmark pdfbookmark = _bookmark[c];
        if (pdfbookmark.title == title) {
          index = c;
          break;
        }
      }
    }
    removeAt(index);
  }

  /// Remove the bookmark from the document at the specified index.
  void removeAt(int index) {
    if (_bookmark == null || _bookmark.length < _list.length) {
      _bookmark = <PdfBookmark>[];
      for (int n = 0; n < _list.length; n++) {
        ArgumentError.checkNotNull(_list[n] as PdfBookmark, 'bookmark');
        _bookmark.add(_list[n] as PdfBookmark);
      }
    }
    if (index < 0 || index >= _bookmark.length) {
      throw RangeError.value(index);
    }
    if (_bookmark[index] is PdfBookmark) {
      final PdfBookmark current = _bookmark[index];
      if (index == 0) {
        if (current._dictionary.containsKey(_DictionaryProperties.next)) {
          _dictionary.setProperty(_DictionaryProperties.first,
              current._dictionary[_DictionaryProperties.next]);
        } else {
          _dictionary.remove(_DictionaryProperties.first);
          _dictionary.remove(_DictionaryProperties.last);
        }
      } else if ((current._parent != null) &&
          (current._previous != null) &&
          (current._next != null)) {
        current._previous._dictionary.setProperty(_DictionaryProperties.next,
            current._dictionary[_DictionaryProperties.next]);
        current._next._dictionary.setProperty(_DictionaryProperties.prev,
            current._dictionary[_DictionaryProperties.prev]);
      } else if ((current._parent != null) &&
          (current._previous != null) &&
          (current._next == null)) {
        current._previous._dictionary.remove(_DictionaryProperties.next);
        current._parent._dictionary.setProperty(_DictionaryProperties.last,
            current._dictionary[_DictionaryProperties.prev]);
      }
      if (current._parent != null) {
        current._parent._list.remove(current);
      }
    }
    _bookmark.clear();
    _updateFields();
  }

  /// Removes all the bookmark from the collection.
  void clear() {
    _list.clear();
    if (_bookmark != null) {
      _bookmark.clear();
    }
    _updateFields();
  }

  /// Inserts a new outline at the specified index.
  PdfBookmark insert(int index, String title) {
    ArgumentError.checkNotNull(title, 'title');
    if (index < 0 || index > count) {
      throw RangeError.value(index);
    }
    PdfBookmark outline;
    if (index == count) {
      outline = add(title);
    } else {
      final PdfBookmark next = this[index];
      final PdfBookmark prevoius = (index == 0) ? null : this[index - 1];
      outline = PdfBookmark._internal(title, this, prevoius, next);
      _list.insert(index, outline);
      if (prevoius != null) {
        prevoius._next = outline;
      }
      next._previous = outline;
      _updateFields();
    }
    return outline;
  }

  /// Updates all outline dictionary fields.
  void _updateFields() {
    if (count > 0) {
      final int newCount = _isExpanded ? _list.length : -_list.length;
      _dictionary[_DictionaryProperties.count] = _PdfNumber(newCount);
      _dictionary.setProperty(
          _DictionaryProperties.first, _PdfReferenceHolder(this[0]));
      _dictionary.setProperty(
          _DictionaryProperties.last, _PdfReferenceHolder(this[count - 1]));
    } else {
      _dictionary.clear();
    }
  }

  /// Gets the element.
  @override
  _IPdfPrimitive get _element => _dictionary;

  set _element(_IPdfPrimitive value) {
    throw ArgumentError();
  }
}
